/*
 * The MIT License
 *
 * Copyright 2021 Sebastian Schauer.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package sudoku.model.board;

import java.util.ArrayList;
import java.util.List;
import sudoku.controller.AreaTransport;

/**
 * A sudoku area. This is a set of cells in which the cells are influencing
 * each other, e.g. a row or a column.
 * 
 * @author Sebastian Schauer
 */
public class Area implements IArea {

    /**
     * A list of all values still available in this area.
     * @return the free values
     */
    public List<Integer> getFreeValues() {
        return freeValues;
    }

    /**
     * The list of all cells belonging to this area.
     */
    private List<Cell> cells;
    /**
     * The list of all remaining free values in this area.
     */
    private List<Integer> freeValues;

    /**
     * Constructor.
     * @param size The size of the sudoku
     */
    public Area(int size) {
        cells = new ArrayList<>();
        freeValues = new ArrayList<>();
        for (int i = 1; i <= size; i++) {
            freeValues.add(i);
        }
        this.foreground = false;
    }
    
    private boolean foreground;
    
    public Area(int size, boolean foreground) {
        cells = new ArrayList<>();
        freeValues = new ArrayList<>();
        for (int i = 1; i <= size; i++) {
            freeValues.add(i);
        }
        this.foreground = foreground;
    }

    /**
     * Checks if 2 cells have the same value.
     * @return True if there are no doublings, false otherwise
     */
    @Override
    public boolean checkConsistency() {
        for (int i = 0; i < cells.size(); i++) {
            if (cells.get(i).getValue() == 0) {
                continue;
            }
            for (int j = i + 1; j < cells.size(); j++) {
                if (cells.get(j).getValue() == 0) {
                    continue;
                }
                if (cells.get(i).getValue() == cells.get(j).getValue()) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Adds a cell, adds the area to the cell.
     * @param cell 
     */
    @Override
    public void addCell(Cell cell) {
        if (!cells.contains(cell)) {
            cells.add(cell);
            cell.addArea(this);
        }
    }

    /**
     * Checks if a value is possible in this area.
     * @param value
     * @return 
     */
    @Override
    public boolean possibleValue(int value) {
        return freeValues.contains(value);
    }

    /**
     * Adds a free value to its list of free values. (After a cell was cleared.)
     * @param val 
     */
    @Override
    public void addFreeValue(int val) {
        if (!possibleValue(val)) {
            freeValues.add(val);
        }
    }

    /**
     * Removes a free value from the list of free values.
     * @param val 
     */
    @Override
    public void removeFreeValue(int val) {
        if (possibleValue(val)) {
            freeValues.remove((Integer) val);
        }
    }

    /**
     * String representation of the area.
     * @return The string representing this area
     */
    @Override
    public String toString() {
        String result = "Area: ";
        for (Cell c : cells) {
            result += c;
        }
        return result;
    }
    
    /**
     * Returns an area transport object (POJO) of this area
     * @return 
     */
    public AreaTransport getAreaTransport() {
        AreaTransport a = new AreaTransport();
        a.foreground = foreground;
        cells.forEach((c) -> {
            a.addCell(c);
        });
        return a;
    }
}
