/*
 * The MIT License
 *
 * Copyright 2021 Sebastian Schauer.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package sudoku.model.board;

import java.util.ArrayList;
import java.util.List;

/**
 * A sudoku cell.
 * 
 * @author Sebastian Schauer
 */
public class Cell {

    /**
     * The list of areas this cell belongs to.
     * @return The list of areas
     */
    public List<IArea> getAreas() {
        return areas;
    }

    /**
     * The actual value of the cell. 0 if no value is set.
     * @return The value, or 0 if no value is set
     */
    public int getValue() {
        return value;
    }
    
    /**
     * Get the x coordinate of the cell.
     * @return The x coordinate of the cell (0-based).
     */
    public int getX() {
        return this.x;
    }
    
    /**
     * Get the y coordinate of the cell.
     * @return The y coordinate of the cell (0-based).
     */
    public int getY() {
        return this.y;
    }

    /**
     * The list of areas this cell is in.
     */
    private List<IArea> areas;
    /**
     * The actual value of the cell, or 0 if nothing is set.
     */
    private int value;
    /**
     * The size of the sudoku (and therefore the number of values the cell can handle).
     */
    private int size;
    /**
     * The 0-based x position of the cell in the sudoku.
     */
    private int x;
    /**
     * The 0-based y position of the cell in the sudoku.
     */
    private int y;
    /**
     * The sudoku the cell belongs to.
     */
    private Sudoku mySudoku;

    /**
     * The constructor of the cell.
     * @param size The size of the sudoku
     * @param x The 0-based x position of the cell in the sudoku
     * @param y The 0-based y position of the cell in the sudoku
     * @param sudoku The sudoku the cell belongs to
     */
    public Cell(int size, int x, int y, Sudoku sudoku) {
        value = 0;
        this.size = size;
        this.mySudoku = sudoku;
        this.x = x;
        this.y = y;
        areas = new ArrayList<>();
    }

    /**
     * Adds an area to the cell and the cell to the area
     * @param area The area the cell is added to
     */
    public void addArea(IArea area) {
        if (!areas.contains(area)) {
            getAreas().add(area);
            area.addCell(this);
        }
    }

    /**
     * Sets a value to the cell. No 0 possible - choose clearValue(). No action
     * if the value is forbidden in this cell.
     * @param val The desired value
     * @return True if the value could be set, false otherwise
     */
    public boolean setValue(int val) {
        if (val < 1 || val > size) {
            return false;
        }
        for (IArea a : areas) {
            if (!a.possibleValue(val)) {
                return false;
            }
        }
        this.value = val;
        for (IArea a : areas) {
            a.removeFreeValue(val);
        }
        mySudoku.getFreeCells().remove(this);
        return true;
    }

    /**
     * Clears the cell, marks the old value available to the areas
     */
    public void clearValue() {
        if (this.value == 0)
            return;
        int oldval = this.value;
        this.value = 0;
        for (IArea a : areas) {
            a.addFreeValue(oldval);
        }
        mySudoku.markCellAsFree(this);
    }

    /**
     * String representation of the cell
     * @return The string representation
     */
    @Override
    public String toString() {
        return "(" + x + "," + y + ":" + value + ")";
    }
}
