/*
 * The MIT License
 *
 * Copyright 2021 Sebastian Schauer.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package sudoku.view;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowStateListener;
import java.io.File;
import static java.lang.Integer.parseInt;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import sudoku.controller.GameController;
import sudoku.controller.SudokuTransport;
import sudoku.model.games.GameType;

/**
 * The main window of the program.
 * 
 * @author Sebastian Schauer
 */
public class Window extends JFrame {
    
    /**
     * The menu bar of the main window
     */
    private final JMenuBar menuBar;
    /**
     * The game controller for this window
     */
    private final GameController myGC;
    /**
     * The solve button
     */
    private final JButton btnSolve;
    /**
     * The JPanel holding the sudoku
     */
    private SudokuView sudokuView;
    /**
     * A map holding a radio button for each game type
     */
    public final Map<GameType, JRadioButton> btnType;
    /**
     * The button group holding the radio buttons for each game type
     */
    public ButtonGroup bgrType;
    /**
     * The last file we read, or null
     */
    private File lastFile;
    /**
     * The panel holding the (radio) buttons
     */
    private final JPanel btnPanel;
    /**
     * The panel holding the sudoku view
     */
    private final JPanel suPanel;
    /**
     * The panel for holding space on the left side
     */
    private final JPanel leftPanel;
    
    
    /**
     * The constructor of the main window
     * @param gc The game controller
     */
    public Window(GameController gc) {
        myGC = gc;
        this.setTitle("SudokuSolver");
        setSize(700, 600);
        
        setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
        
        lastFile = null;
        
        setLayout(new BorderLayout());
        btnPanel = new JPanel();
        btnPanel.setLayout(new BoxLayout(btnPanel, BoxLayout.Y_AXIS));
        btnPanel.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));
        
        this.addComponentListener(new ComponentAdapter() {
            @Override
            public void componentResized(ComponentEvent arg0) {
                resizeSudoku();
            }            
        });
        this.addWindowStateListener(new WindowStateListener() {
            @Override
            public void windowStateChanged(WindowEvent arg0) {
                revalidate();       
                resizeSudoku();
            }
        });
        this.addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent arg0) {
                clickedExit(null);
            }
        });
        
        // MENU
        menuBar = new JMenuBar();
        
        JMenu file = new JMenu("File");
        
        JMenu mnuNew = new JMenu("New");
        for(int i : new int[] {4, 9, 16, 25}) {
            JMenuItem mnuNewX = new JMenuItem(i + "x" + i);
            mnuNewX.setActionCommand("" + i);
            mnuNewX.addActionListener(a -> clickedNew(a));
            mnuNew.add(mnuNewX);
        }       
        file.add(mnuNew);
        
        JMenuItem mnuLoad = new JMenuItem("Load...");
        mnuLoad.addActionListener(a -> clickedLoad(a));
        file.add(mnuLoad);
        
        JMenuItem mnuSave = new JMenuItem("Save");
        mnuSave.addActionListener(a -> clickedSave(a));
        file.add(mnuSave);
        
        JMenuItem mnuSaveAs = new JMenuItem("Save as...");
        mnuSaveAs.addActionListener(a -> clickedSaveAs(a));
        file.add(mnuSaveAs);
        
        JMenuItem mnuQuit = new JMenuItem("Quit");
        mnuQuit.addActionListener(a -> clickedExit(a));
        file.add(mnuQuit);
        
        menuBar.add(file);
        
        JMenu game = new JMenu("Game");
        JMenuItem solve = new JMenuItem("Solve Game");
        solve.addActionListener(a -> clickedSolve(a));
        game.add(solve);
        menuBar.add(game);
        
        JMenu help = new JMenu("Help");
        JMenuItem mnuHelp = new JMenuItem("Usage");
        mnuHelp.addActionListener(a -> clickedUsage(a));
        help.add(mnuHelp);
        menuBar.add(help);
        
        this.setJMenuBar(menuBar);
        
        // RADIO BUTTONS
        bgrType = new ButtonGroup();
        btnType = new HashMap<>();
        int i = 0;
        for (GameType type : GameType.values()) {
            JRadioButton radio = new JRadioButton(type.name());
            radio.setActionCommand(type.name());
            radio.addActionListener(a -> clickedType(a));
            if (i == 0)
                radio.setEnabled(true);
            bgrType.add(radio);
            btnType.put(type, radio);
            btnPanel.add(radio);
            i++;
        }
        
        // BUTTONS
        btnSolve = new JButton("Solve");
        btnSolve.setBounds(550, 450, 130, 80);
        btnSolve.addActionListener(a -> clickedSolve(a));
        btnPanel.add(btnSolve);
        
        getContentPane().add(btnPanel, BorderLayout.EAST);
        
        suPanel = new JPanel(null);
        getContentPane().add(suPanel, BorderLayout.CENTER);
        
        leftPanel = new JPanel();
        leftPanel.setPreferredSize(new Dimension(20,20));
        getContentPane().add(leftPanel, BorderLayout.WEST);
        
        setVisible(true);
    }
    
    /**
     * Displays the given sudoku
     * @param st 
     */
    public void setSudoku(SudokuTransport st) {
        if (sudokuView != null)
            suPanel.remove(sudokuView);
        sudokuView = new SudokuView(st, this);
        sudokuView.setSize(suPanel.getSize());
        suPanel.add(sudokuView);
        btnType.get(st.type).setEnabled(true);
        revalidate();
    }
    /**
     * Resizes the panel holding the sudoku
     */
    private void resizeSudoku() {
        if (sudokuView != null) {
            int width = getContentPane().getWidth() - btnPanel.getWidth() - leftPanel.getWidth();
            int height = getContentPane().getHeight();
            if (width < height) {
                suPanel.setSize(width, width);
                sudokuView.setSize(width - 1, width - 1);
            }
            else {
                suPanel.setSize(height, height);
                sudokuView.setSize(height - 1, height - 1);
            }
        }
    }
    
    /**
     * Get the game type of the displayed sudoku
     * @return The game type
     */
    public GameType getSudokuType() {
        
        for (Entry<GameType, JRadioButton> e : btnType.entrySet()) {
            if (e.getValue().isSelected())
                return e.getKey();
        }
        return GameType.Regular;
    }
    
    private void clickedNew(ActionEvent arg0) {
        if (sudokuView.mayClose()) {
            int size = parseInt(arg0.getActionCommand());
            GameType type = getSudokuType();
            setSudoku(myGC.newSudoku(size, type));
        }
    }
    
    private void clickedExit(ActionEvent arg0) {
        if (!sudokuView.mayClose())
            clickedSave(null);
        else
            System.exit(0);
    }
    
    private void clickedSolve(ActionEvent ae) {
        myGC.solve(sudokuView.getSudokuTransport());
    }
    
    private void clickedLoad(ActionEvent ae) {
        SudokuTransport oldSt = sudokuView.getSudokuTransport();
        if (sudokuView.mayClose()) {
            SudokuTransport newSt = myGC.load(lastFile);
            if (newSt != null)
                setSudoku(newSt);
        }
    }
    
    private void clickedSave(ActionEvent ae) {
        if (lastFile == null)
            clickedSaveAs(ae);
        else
            myGC.write(sudokuView.getSudokuTransport(), lastFile);
    }
    
    private void clickedSaveAs(ActionEvent ae) {
        File f = myGC.save(sudokuView.getSudokuTransport(), lastFile);
        if (f != null)
            lastFile = f;
    }
    
    private void clickedUsage(ActionEvent ae) {
        
    }
    
    public boolean closeSudoku() {
        if (sudokuView != null) {
            if (sudokuView.mayClose()) {
                suPanel.remove(sudokuView);
                sudokuView = null;
                return true;
            }
            else
                return false;
        }
        return true;
    }
    
    public void clickedType(ActionEvent ae) {
        sudokuView.setType(ae.getActionCommand());
    }
}
